from kvix import Context
from kvix.plugin.impl import FromModules
from kvix.plugin import builtin, discovery


class Plugin(FromModules):
    def __init__(self, context: Context):
        FromModules.__init__(self, context, builtin, discovery)



from __future__ import annotations

import locale
from typing import cast
from kvix.util import ensure_key


def get_current_locale() -> str:
    return cast(str, locale.getlocale()[0])


class Text:
    def __init__(self, key: str, default: str | None = None, **l10ns: str):
        self.key = key
        self.default = default or key
        self.l10ns = l10ns

    def setup(self, default: str | None = None, **kwargs: str) -> Text:
        self.default = default or self.key
        self.l10ns.update(kwargs)
        return self

    def __str__(self) -> str:
        return self[get_current_locale()]

    def __getitem__(self, locale: str) -> str:
        return self.l10ns.get(locale, self.default)

    def values(self) -> set[str]:
        return set([self.default, *self.l10ns.values()])


_texts: dict[str, Text] = {}


def gettext(key: str, default: str | None = None, **l10ns: str) -> Text:
    return ensure_key(_texts, key, lambda: Text(key, default, **l10ns))


_ = gettext


class Scope:
    def __init__(self, key: str):
        self._key = key

    def gettext(self, key: str) -> Text:
        return gettext(self._key + "." + key)


_scopes: dict[str, Scope] = {}


def scope(key: str) -> Scope:
    return ensure_key(_scopes, key, lambda: Scope(key))


def _test():
    from kvix.l10n import _

    txt = _("Hello").setup(ru_RU="Привет", de_DE="Hallo")
    print(txt)



from kvix import Context
from kvix.plugin.impl import FromModules

from . import (
    add_action,
    base64decode,
    base64encode,
    calculator,
    edit_action,
    machinist,
    openurl,
    quit,
    settings,
    shell,
    websearch,
)


class Plugin(FromModules):
    def __init__(self, context: Context):
        FromModules.__init__(
            self,
            context,
            add_action,
            base64decode,
            base64encode,
            calculator,
            edit_action,
            machinist,
            openurl,
            quit,
            settings,
            shell,
            websearch,
        )



from kvix.util import Propty


def test_x():
    class T:
        x = Propty(str)

    t = T()
    t.x = "1"
    assert t._x == "1"
    assert t.x == "1"


def test_default_value():
    class T:
        x = Propty(str, default_value="123")

    t = T()
    assert not hasattr(t, "_x")
    assert t.x == "123"
    assert t._x == "123"



from __future__ import annotations

from typing import Any, Callable, Sequence, cast, TypeVar, Generic, Type

from kvix.stor import Stor


T = TypeVar("T")


class Conf:
    def __init__(self):
        self._scopes: dict[str, Scope] = {}
        self._items: dict[str, Item[Any]] = {}

    def _get_data(self) -> dict[Any, Any]:
        ...

    def load(self):
        ...

    def save(self):
        ...

    def scope(self, key: str, title: str | None = None) -> Conf:
        if key in self._items:
            raise RuntimeError('cannot create scope "' + key + '": there is item with such key')
        result = self._scopes.get(key)
        if not result:
            result = Scope(self, key, title)
            self._scopes[key] = result
        return result

    def scopes(self) -> dict[str, Scope]:
        return {**self._scopes}

    def item(self, key: str, type: Type[T] = Type[Any]) -> Item[T]:
        if key in self._scopes:
            raise RuntimeError('cannot create item "' + key + '": there is scope with such key')
        result = self._items.get(key)
        if not result:
            result = Item(self, key, type)
            self._items[key] = result
        return result

    def items(self) -> dict[str, Item[Any]]:
        return {**self._items}


class Scope(Conf):
    def __init__(self, parent: Conf, key: str, title: str | None = None):
        super().__init__()
        self.parent = parent
        self.key = key
        self.title = title

    def setup(self, title: str | None = None):
        self.title = title or self.key
        return self

    def _get_data(self) -> dict[Any, Any]:
        result: Any = self.parent._get_data().setdefault(self.key, {})
        if not isinstance(result, dict):
            raise RuntimeError(
                "dict expected at scope "
                + self.key
                + (" (" + self.title + ")" if self.title else "")
            )
        return cast(dict[Any, Any], result)

    def load(self):
        self.parent.load()

    def save(self):
        self.parent.save()


class StorConf(Conf):
    def __init__(self, stor: Stor):
        Conf.__init__(self)
        self._stor: Stor = stor

    def _get_data(self) -> dict[Any, Any]:
        result = self._stor.data
        if not isinstance(result, dict):
            raise RuntimeError("stor.data expected to be dict)")
        return cast(dict[Any, Any], result)

    def load(self):
        self._stor.load()

    def save(self):
        self._stor.save()


class Item(Generic[T]):
    def __init__(self, parent: Conf, key: str, type: Type[T] = Type[Any]):
        self.parent = parent
        self.key = key
        self.type = type
        self.setup()

    def setup(
        self,
        title: str | None = None,
        default: Any = None,
        read_mapping: Callable[[Any], Any] | None = None,
        enum: Sequence[Any] | None = None,
        on_change: Callable[[Any], None] = lambda value: None,
    ):
        self._title = title
        self._default = default
        self._read_mapping = read_mapping
        self._enum = enum
        self._on_change = on_change
        return self

    def read(self):
        result = self.parent._get_data().get(self.key)
        if self._default and not result:
            result = self._default
        if self._read_mapping:
            result = self._read_mapping(result)
        # todo handle enum somehow?
        return result

    def write(self, value: Any):
        data = self.parent._get_data()
        old_value = data.get(self.key, self._default)
        if old_value != value:
            data[self.key] = value
            if value == self._default:
                del data[self.key]
            self._on_change(value)



import http.client
import http.server
import threading
from typing import Callable
from kvix import Context

from kvix.conf import Conf
from kvix.l10n import _

listen_host_conf_title = _("Host").setup(ru_RU="Хост")
listen_port_conf_title = _("Port").setup(ru_RU="Порт")
remote_scope_title_text = _("Remote Control").setup(ru_RU="Удалённое управление")


class Configurable:
    def __init__(self, conf: Conf):
        self._conf = conf

    def _scope(self) -> Conf:
        return self._conf.scope("kvix").scope("remote").setup(title=str(remote_scope_title_text))

    def _host(self) -> str:
        item = (
            self._scope()
            .item("host")
            .setup(
                default="127.0.0.1",
                title=str(listen_host_conf_title),
                on_change=self._on_change_conf,
            )
        )
        return str(item.read())

    def _port(self) -> int:
        item = (
            self._scope()
            .item("port")
            .setup(
                default="23844",
                title=str(listen_port_conf_title),
                read_mapping=int,
                on_change=self._on_change_conf,
            )
        )
        return int(str(item.read()))

    def _on_change_conf(self) -> None:
        pass


def apply(msg, func, *args, **kwargs):
    try:
        return apply(func, *args, **kwargs)
    except Exception as e:
        raise RuntimeError(msg) from e


class Client(Configurable):
    def is_server_ready(self):
        try:
            self._http("/ping")
            return True
        except:
            return False

    def ping(self):
        self._http("/ping")

    def activate(self):
        self._http("/activate")

    def quit(self):
        self._http("/quit")

    def _http(self, path: str):
        conn = http.client.HTTPConnection(self._host() + ":" + str(self._port()))
        try:
            conn.request("POST", path)
            status = conn.getresponse().status
            body = conn.getresponse().read()
            if 2 != status / 100:
                raise RuntimeError(
                    "remote call %s failed: status %d, body %s" % (path, status, body)
                )
        except Exception as e:
            raise RuntimeError("remote call to %s failed: %s" % (path, e)) from e


class Server(Configurable):
    def __init__(self, context: Context, on_activate: Callable[[], None]):
        Configurable.__init__(self, context.conf)
        self._context = context
        self._on_activate = on_activate
        self._http_server = None

    def run(self):
        "run in this thread"
        this_server = self

        class Handler(http.server.BaseHTTPRequestHandler):
            def do_POST(self):
                if self.path == "/ping":
                    self.response(200, "ok")
                elif self.path == "/activate":
                    try:
                        this_server._on_activate()
                        self.response(200, "ok")
                    except Exception as e:
                        self._error(e)
                elif self.path == "/quit":
                    try:
                        this_server._context.quit()
                        self.response(200, "ok")
                    except Exception as e:
                        self._error(e)
                else:
                    self.response(400, "wrong request")

            def _error(self, e: Exception | None = None):
                if e:
                    print(e)
                self.response(500, "Internal server error, see server logs for details.")

            def response(self, status: int, msg: str):
                self.send_response(status)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self.wfile.write(b'{"msg":"' + bytes(msg, "UTF-8") + b'"}')

        self._http_server = http.server.HTTPServer((self._host(), int(self._port())), Handler)
        self._http_server.serve_forever()

    def start(self):
        "start server in background"
        self.stop()
        threading.Thread(name="kvix.remote.Server", target=self.run).start()

    def stop(self):
        "stop server"
        if self._http_server:
            self._http_server.shutdown()
            self._http_server = None

    def _on_change_conf(self) -> None:
        self.start()



from __future__ import annotations

from inspect import isclass
from types import ModuleType
from typing import Any, Callable, Protocol, cast, Sequence

import funcy
import re

import kvix
from kvix import Action, ActionType, Context, Item, ItemAlt, Ui
from kvix.l10n import _
from kvix.stor import Stor
from kvix.util import Propty, query_match, apply_template
from kvix.util import ThreadRouter

unnamed_text = _("Unnamed").setup(ru_RU="Без названия", de_DE="Ohne Titel")
execute_text = _("Execute").setup(ru_RU="Выполнить", de_DE="Ausführen")
title_text = _("Title").setup(ru_RU="Название", de_DE="Bezeichnung")
description_text = _("Description").setup(ru_RU="Описание", de_DE="Beschreibung")
pattern_text = _("Regular expression").setup(
    ru_RU="Регулярное выражение", de_DE="Regular expression"
)
ok_text = _("OK")
cancel_text = _("Cancel").setup(ru_RU="Отмена", de_DE="Abbrechen")


class WithTitleStr:
    def __init__(self, title: Any):
        self._title = title

    def __str__(self) -> str:
        return str(self._title)


class BaseItemAlt(kvix.ItemAlt, WithTitleStr):
    def __init__(self, title: Any, command: Callable[[], None]):
        WithTitleStr.__init__(self, title)
        self._command = command

    def execute(self):
        return self._command()


class BaseItem(kvix.Item, WithTitleStr):
    def __init__(self, title: Any, alts: Sequence[kvix.ItemAlt]):
        WithTitleStr.__init__(self, title)
        self._alts = alts


class BaseItemSource(kvix.ItemSource):
    def __init__(self, func: Callable[[str], list[kvix.Item]]):
        self._func = func

    def search(self, query: str) -> Sequence[Item]:
        return self._func(query)


class EmptyItemSource(kvix.ItemSource):
    def search(self, query: str) -> Sequence[kvix.Item]:
        return []


class BaseSelector(kvix.Selector):
    title = Propty(default_value=unnamed_text)
    item_source = Propty(kvix.ItemSource, default_supplier=lambda: EmptyItemSource())

    def __init__(
        self,
        item_source: kvix.ItemSource = EmptyItemSource(),
        title: str | None = None,
    ):
        self.item_source = item_source
        if title is not None:
            self.title = title


class ActionFactory(Protocol):
    def __call__(
        self,
        action_type: ActionType,
        title: str,
        description: str = "",
        pattern: str = "",
        **config: Any,
    ) -> Action:
        ...


class BaseActionType(ActionType):
    BUILTIN_PARAMS = ["title", "description", "pattern"]

    def __init__(
        self,
        context: kvix.Context,
        id: str,
        title: str,
        action_factory: ActionFactory | None = None,
        config_entry_texts: dict[str, Any] = {},
    ):
        self.context = context
        self.id = id
        self.title = title
        self.action_factory = action_factory
        self.config_entry_texts = config_entry_texts

    def create_default_action(
        self, title: str, description: str = "", pattern: str = "", **params: Any
    ) -> Action:
        return cast(ActionFactory, self.action_factory)(
            self, title=title, description=description, pattern=pattern, **params
        )

    def action_from_config(self, value: Any) -> Action:
        dic = {**self._assert_config_valid(value)}
        return self.create_default_action(
            title=str(dic["title"]),
            description=str(dic.get("description", "")),
            pattern=str(dic.get("pattern", "")),
            **funcy.omit(dic, ["action_type"] + BaseActionType.BUILTIN_PARAMS),
        )

    def _assert_config_valid(self, value: Any) -> dict[Any, Any]:
        if not isinstance(value, dict):
            raise RuntimeError("json must be object, " + value + " given")
        value = cast(dict[Any, Any], value)
        if "type" not in value:
            raise RuntimeError('"type" must be in json object')
        if value.get("type") != self.id:
            raise RuntimeError(
                "wrong type got " + str(value.get("type")) + ", expected " + self.id
            )
        return value

    def create_editor(self, builder: kvix.DialogBuilder) -> None:
        builder.create_entry("title", str(title_text))
        builder.create_entry("description", str(description_text))
        builder.create_entry("pattern", str(pattern_text))
        for key, text in self.config_entry_texts.items():
            builder.create_entry(key, str(text))

        def load(value: Any | None):
            if isinstance(value, Action):
                builder.widget("title").set_value(value.title)
                builder.widget("description").set_value(value._description)
                builder.widget("pattern").set_value(value._pattern)
                for key in self.config_entry_texts:
                    builder.widget(key).set_value(value._config[key])

        builder.on_load(load)

        def save(value: Any | None) -> Any:
            if isinstance(value, Action):
                value.title = builder.widget("title").get_value()
                value._description = builder.widget("description").get_value()
                value._pattern = builder.widget("pattern").get_value()
                for key in self.config_entry_texts:
                    value._config[key] = builder.widget(key).get_value()
            elif isinstance(value, dict):
                value["title"] = builder.widget("title").get_value()
                value["description"] = builder.widget("description").get_value()
                value["pattern"] = builder.widget("pattern").get_value()
                for key in self.config_entry_texts:
                    value[key] = builder.widget(key).get_value()
            elif not value:
                value = self.create_default_action(
                    builder.widget("title").get_value(),
                    builder.widget("description").get_value(),
                    builder.widget("pattern").get_value(),
                    **{
                        key: builder.widget(key).get_value()
                        for key in self.config_entry_texts.keys() - BaseActionType.BUILTIN_PARAMS
                    },
                )
            return value

        builder.on_save(save)


class BaseAction(Action):
    def __init__(
        self,
        action_type: ActionType,
        title: str,
        description: str = "",
        pattern: str = "",
        **params: Any,
    ):
        self._action_type = action_type
        self._title = title
        self._description = description or title
        self._pattern = pattern
        self._set_params(**params)

    def _set_params(self, **params: Any) -> None:
        self._config = {**params}  # todo rename to params
        self._on_after_set_params(**params)

    def _on_after_set_params(self, **params: Any) -> None:
        pass

    def search(self, query: str) -> Sequence[kvix.Item]:
        if not self._match(query):
            return []
        return self._create_items(query)

    def _match(self, query: str) -> bool:
        if not query:
            return True
        if self._pattern:
            return re.compile(self._pattern).match(query) and True or False
        return query_match(query or "", *self._word_list()) and True or False

    def _word_list(self) -> Sequence[str]:
        return [self.title, self._description, *self._config.values()]

    def _create_items(self, query: str) -> Sequence[Item]:
        return (self._create_single_item(query),)

    def _create_single_item(self, query: str) -> Item:
        return BaseItem(self._get_single_item_title(query), self._create_single_item_alts(query))

    def _get_single_item_title(self, query: str) -> str:
        return apply_template(self._title, query=query)

    def _create_single_item_alts(self, query: str) -> Sequence[ItemAlt]:
        return (self._create_single_item_single_alt(query),)

    def _create_single_item_single_alt(self, query: str) -> ItemAlt:
        return BaseItemAlt(execute_text, lambda: self._run(query=query))

    def _run(self, query: str) -> None:
        raise NotImplementedError()

    def to_config(self) -> dict[str, Any]:
        return {
            "type": self.action_type.id,
            "title": self.title,
            "description": self._description,
            "pattern": self._pattern,
            **self._config,
        }


class BaseActionRegistry(kvix.ActionRegistry):
    def __init__(self, stor: Stor):
        self.stor: Stor = stor

    def load(self):
        self.stor.load()
        self._actions: list[Action] = []
        for action_config in self.stor.data or []:
            self._actions.append(self.action_from_config(action_config))

    def save(self) -> None:
        self.stor.data = [action.to_config() for action in self.actions]
        self.stor.save()

    def add_action_type(self, action_type: ActionType) -> None:
        if action_type.id in self.action_types:
            raise RuntimeError("duplicate action type id=" + action_type.id)
        self.action_types[action_type.id] = action_type

    def action_from_config(self, value: Any) -> Action:
        if not isinstance(value, dict):
            raise RuntimeError("dict expected, got " + type(value))
        value = cast(dict[Any, Any], value)
        if "type" not in value:
            raise RuntimeError('"type" expected')
        type_id = value["type"]
        if not isinstance(type_id, str):
            raise RuntimeError('"type" expected to be str, got ' + type(type_id))
        if type_id not in self.action_types:
            raise RuntimeError("uknown action type id=" + type_id)
        action_type: ActionType = self.action_types[type_id]
        return action_type.action_from_config(value)

    def search(self, query: str) -> Sequence[kvix.Item]:
        result: list[kvix.Item] = []
        for action in self.actions:
            for item in action.search(query):
                result.append(item)
        return result


class BasePlugin(kvix.Plugin):
    def __init__(self, context: kvix.Context):
        self.context = context

    def get_action_types(self) -> Sequence[ActionType]:
        self._single_action_type = self._create_single_action_type()
        return [self._single_action_type]

    def _create_single_action_type(self) -> ActionType:
        raise NotImplementedError()

    def get_actions(self) -> Sequence[Action]:
        return []


class FromModule(BasePlugin):
    def __init__(self, context: Context, module: ModuleType):
        BasePlugin.__init__(self, context)
        self._wrap = self._create_plugin(module)

    def _create_plugin(self, module: ModuleType) -> kvix.Plugin | None:
        if not hasattr(module, "Plugin"):
            return None
        PluginClass = getattr(module, "Plugin")
        if not isclass(PluginClass):
            return None
        if not issubclass(PluginClass, kvix.Plugin):
            return None
        return PluginClass(self.context)

    def get_action_types(self) -> Sequence[ActionType]:
        return self._wrap.get_action_types() if self._wrap else []

    def get_actions(self) -> Sequence[Action]:
        return self._wrap.get_actions() if self._wrap else []


class BaseUi(Ui):
    def __init__(self):
        self._thread_router: ThreadRouter = cast(ThreadRouter, None)
        self._on_ready_listeners: list[Callable[[], None]] = []

    def on_ready(self, f: Callable[[], None]) -> None:
        self._on_ready_listeners.append(f)

    def _call_on_ready_listeners(self):
        for f in self._on_ready_listeners:
            f()

    def run(self):
        self._thread_router = ThreadRouter()

    def _exec_in_mainloop(self, func: Callable[[], None]) -> None:
        self._thread_router.exec(func)

    def _process_mainloop(self):
        self._thread_router.process()



from typing import Any, Callable, Sequence, cast, Tuple

import pkg_resources
import pyclip
from PIL import Image, ImageTk

import kvix
import kvix.impl
import kvix.ui
import kvix.ui.tk.uni.classic as uni
from kvix import Item, ItemAlt, ItemSource
from kvix.conf import Conf
from kvix.impl import BaseSelector, BaseUi, EmptyItemSource, cancel_text, ok_text
from kvix.l10n import _

from .util import find_all_children

style_config_item_title_text = _('Theme ("dark" or "light")').setup(
    ru_RU='Тема ("dark" или "light")'
)


def get_logo():
    return Image.open(pkg_resources.resource_filename("kvix", "logo.jpg"))


def run_periodically(root: uni.Tk, interval_ms: int, action):
    def on_timer():
        root.after(interval_ms, on_timer)
        action()

    on_timer()


class Ui(BaseUi):
    def __init__(self, conf: Conf):
        super().__init__()
        self.root = uni.Tk()
        self.root.wm_iconphoto(False, ImageTk.PhotoImage(get_logo()))
        self.root.title("kvix!")
        self.root.withdraw()

        style_conf_item = (
            conf.scope("ui")
            .scope("tk", "Tk")
            .item("theme")
            .setup(
                title=str(style_config_item_title_text), default="light", on_change=self._set_theme
            )
        )
        self._style = uni.Style(self.root)
        self._style.theme_use(str(style_conf_item.read()))

    def _set_theme(self, theme_name: str) -> None:
        self._style.theme_use(theme_name)

    def run(self):
        BaseUi.run(self)
        run_periodically(self.root, 10, self._process_mainloop)
        self.root.after(0, self._call_on_ready_listeners)
        self.root.mainloop()

    def _exec_in_mainloop(self, func: Callable[[], None]) -> None:
        self._thread_router.exec(func)

    def selector(self) -> kvix.Selector:
        return Selector(self)

    def dialog(self, create_dialog: Callable[[kvix.DialogBuilder], None]) -> None:
        return Dialog(self, create_dialog)

    def destroy(self):
        self._exec_in_mainloop(self._do_destroy)

    def _do_destroy(self):
        self.root.destroy()

    def copy_to_clipboard(self, data: bytes, format_id: str = "text/plain;charset=utf-8") -> None:
        self._assert_supported_clipboard_forman_id(format_id)
        try:
            pyclip.copy(data.decode() if data else None)
        except Exception as e:
            print("error copying to clipboard: ", e)

    def _assert_supported_clipboard_forman_id(self, format_id: str) -> None:
        _SUPPORTED_CLIPBOARD_FORMATS = set(
            [
                "STRING",
                "UTF8_STRING",
                "TEXT",
                "text/plain",
                "text/plain;charset=utf-8",
            ]
        )
        if format_id not in _SUPPORTED_CLIPBOARD_FORMATS:
            raise Exception("unsupported clipboard format id: " + format_id)

    def paste_from_clipboard(self, format_id: str = "text/plain;charset=utf-8") -> bytes:
        self._assert_supported_clipboard_forman_id(format_id)
        try:
            result = pyclip.paste()
            if result is None:
                return result
            elif isinstance(result, bytes):
                return cast(bytes, result)
            else:
                return str(result).encode()
        except Exception as e:
            print("error pasting from clipboard: ", e)

    def hide(self) -> None:
        def go():
            for widget in find_all_children(self.root):
                if "Toplevel" == widget.winfo_class():
                    cast(uni.Toplevel, widget).withdraw()

        self._exec_in_mainloop(go)


class ModalWindow:
    def __init__(self, parent: Ui):
        self.parent = parent
        self.title = "kvix"
        self._create_window()

    def _create_window(self):
        self._window = uni.Toplevel(self.parent.root)
        # self._widow.wm_iconphoto(False, Imageuni.PhotoImage(get_logo()))
        # self._window.transient(self.parent.root)
        self._window.title(self.title)
        self._window.geometry("500x500")
        self._center_on_window(500, 500)
        self._window.columnconfigure(0, weight=1)
        self._window.rowconfigure(0, weight=1)
        self._window.bind("<Escape>", lambda _: self._do_hide())
        self._window.protocol("WM_DELETE_WINDOW", lambda: self._do_hide())
        # self._window.bind("<FocusOut>", lambda _: self._do_hide())
        self._window.withdraw()

    def _center_on_window(self, width: int = 500, height: int = 500):
        screen_width = self._window.winfo_screenwidth()
        screen_height = self._window.winfo_screenheight()
        x = (screen_width / 2) - (width / 2)
        y = (screen_height / 2) - (height / 2)
        self._window.geometry("%dx%d+%d+%d" % (width, height, x, y))

    def show(self):
        self.parent._exec_in_mainloop(self._do_show)

    def _do_show(self):
        self._maybe_hide_before_show()
        self._window.deiconify()
        # todo: _tkinter.TclError: bad window path name ".!toplevel"
        self._window.title(self.title)
        # do our best efforts to bring window to foreground and give it focus
        self._window.focus_set()
        self.parent.root.lift()
        self.parent.root.attributes("-topmost", True)
        self.parent.root.after_idle(self.parent.root.attributes, "-topmost", False)
        self.parent.root.focus_force()

    def _maybe_hide_before_show(self):
        # there is a dilemma, shoud we hide before show, or not
        # if we do, we get window blinking on reactivate
        # if we do not, the window is not brought to foreground if already shown in background
        pass  # self._do_hide()

    def hide(self) -> None:
        self.parent._exec_in_mainloop(self._do_hide)

    def _do_hide(self) -> None:
        self._window.withdraw()

    def destroy(self):
        if not self._window:
            return

        def f():
            if not self._window:
                return
            self._window.destroy()
            self._window = None

        self.parent._exec_in_mainloop(f)


class Selector(ModalWindow, BaseSelector):
    actions = []

    def __init__(self, parent: Ui, item_source: ItemSource = EmptyItemSource()):
        ModalWindow.__init__(self, parent)
        kvix.impl.BaseSelector.__init__(self, item_source)
        self.result: Item | None = None
        self._init_window()

    def _init_window(self):
        self._window.bind("<Return>", cast(Any, lambda x: self._on_enter(0)))
        self._window.bind("<Alt-KeyPress-Return>", cast(Any, lambda x: self._on_enter(1)))
        for key in ["<Menu>", "Shift-F10"]:  # menu key: name=menu, value=65383
            self._window.bind(key, lambda x: self._on_popup_key_press())

        self._mainframe = uni.Frame(self._window)
        self._mainframe.grid(column=0, row=0, sticky="nsew")
        self._mainframe.rowconfigure(1, weight=1)
        self._mainframe.columnconfigure(0, weight=1)

        self._search_query = uni.StringVar()
        self._search_query.trace_add("write", lambda *_: self._on_query_entry_type())
        self._search_entry = uni.Entry(self._mainframe, textvariable=self._search_query)
        self._search_entry.grid(column=0, row=0, sticky="ew")

        self._search_entry.bind("<Up>", lambda _: self._on_search_entry_key_arrow_up())
        self._search_entry.bind("<Down>", lambda _: self._on_search_entry_key_arrow_down())
        self._search_entry.bind("<Prior>", lambda _: self._on_search_entry_key_page_up())
        self._search_entry.bind("<Next>", lambda _: self._on_search_entry_key_page_down())
        for key in ("<Home>", "<Control-Home>"):
            self._search_entry.bind(key, lambda _: self._on_search_entry_key_ctrl_home())
        for key in ("<End>", "<Control-End>"):
            self._search_entry.bind(key, lambda _: self._on_search_entry_key_ctrl_end())

        result_frame = uni.Frame(self._mainframe)
        result_frame.rowconfigure(0, weight=1)
        result_frame.columnconfigure(0, weight=1)
        result_frame.grid(column=0, row=1, sticky="nsew")

        self._result_list = uni.StringVar()
        self._result_listbox = uni.Listbox(
            result_frame,
            listvariable=self._result_list,
            takefocus=False,
            selectmode="browse",
        )
        self._result_listbox.grid(column=0, row=0, sticky="nsew")
        self._result_listbox.bind("<Button-1>", self._on_list_left_click)
        self._result_listbox.bind("<Button-3>", self._on_list_right_click)

        scrollbar = uni.Scrollbar(result_frame, orient="vertical")
        scrollbar.grid(column=1, row=0, sticky="nsew")

        # bind scrollbar to result list
        self._result_listbox.config(yscrollcommand=scrollbar.set)
        scrollbar.configure(command=self._result_listbox.yview)

        self._on_query_entry_type()

    def _hide(self, event):
        self._window.withdraw()

    def _on_enter(self, alt_index: int = -1):
        if 0 == len(self._item_list):
            return
        item: Item | None = self._get_selected_item()
        if not item:
            item = self._item_list[0]
        alts: list[ItemAlt] = item.alts
        if alt_index >= 0 and alt_index < len(alts):
            self._maybe_hide_on_execute_item_alt()
            alts[alt_index].execute()

    def _maybe_hide_on_execute_item_alt(self):
        pass  # self.hide() # prefer not to autohide selector, delegate this to actons

    def _get_selected_item(self) -> Item | None:
        index = self._get_selected_index()
        if isinstance(index, int):
            return self._item_list[index]

    def _set_selected_index(self, index: int | None):
        if 0 == self._result_listbox.size():
            return
        selection: Tuple[int, int] | Tuple[int] | Tuple[()] = self._result_listbox.curselection()
        if selection:
            self._result_listbox.selection_clear(*selection)
        if index is not None and index >= 0 and index < self._result_listbox.size():
            self._result_listbox.selection_set(index)
            self._result_listbox.see(index)

    def _get_selected_index(self) -> int | None:
        selection: Tuple[int, int] | Tuple[()] = self._result_listbox.curselection()
        if not selection:
            return None
        return selection[0]

    def _on_search_entry_key_arrow_up(self):
        self._set_selected_index((self._get_selected_index() or self._result_listbox.size()) - 1)

    def _on_search_entry_key_arrow_down(self):
        index = self._get_selected_index()
        if index is not None and index < self._result_listbox.size() - 1:
            self._set_selected_index(index + 1)
        else:
            self._set_selected_index(0)

    def _on_search_entry_key_page_up(self):
        index = self._get_selected_index()
        if index is not None:
            self._set_selected_index(max(index - 10, 0))
        else:
            self._set_selected_index(0)

    def _on_search_entry_key_page_down(self):
        index = self._get_selected_index()
        if index is not None:
            self._set_selected_index(min(index + 10, self._result_listbox.size() - 1))
        else:
            self._set_selected_index(0)

    def _on_search_entry_key_ctrl_home(self):
        self._set_selected_index(0)

    def _on_search_entry_key_ctrl_end(self):
        self._set_selected_index(self._result_listbox.size() - 1)

    def _on_list_left_click(self, event):
        self._select_item_at_y_pos(event.y)
        self._on_enter(0)

    def _on_list_right_click(self, event):
        self._select_item_at_y_pos(event.y)
        item: Item | None = self._get_selected_item()
        if item:
            self._show_popup_menu(item, event.x_root, event.y_root)

    def _show_popup_menu(self, item: Item, x: int, y: int) -> None:
        alts: list[ItemAlt] = item.alts
        if len(alts) >= 0:
            popup_menu = uni.Menu(self._result_listbox, tearoff=0)
            popup_menu.bind("<FocusOut>", lambda event: popup_menu.destroy())
            for alt in alts:

                def execute(alt: ItemAlt = alt):
                    popup_menu.destroy()
                    self._maybe_hide_on_execute_item_alt()
                    self.parent.root.after_idle(lambda: alt.execute())

                popup_menu.add_command(label=str(alt), command=execute)
            popup_menu.tk_popup(x, y)

    def _on_popup_key_press(self):
        item: Item | None = self._get_selected_item()
        if item:
            x: int = int(
                (
                    int(self._result_listbox.winfo_rootx())
                    + int(self._result_listbox.winfo_width()) / 3
                )
            )

            selected_index: int | None = self._get_selected_index()
            if isinstance(selected_index, int):
                coords: tuple[int, int, int, int] = self._result_listbox.bbox(selected_index)
                y: int = int(self._result_listbox.winfo_rooty() + (coords[1] + coords[3] / 2))
            else:
                y: int = int(
                    (
                        int(self._result_listbox.winfo_rooty())
                        + int(self._result_listbox.winfo_height()) / 3
                    )
                )
            self._show_popup_menu(item, x, y)

    def _select_item_at_y_pos(self, y: int):
        self._result_listbox.selection_clear(0, uni.END)
        self._result_listbox.selection_set(self._result_listbox.nearest(y))

    def activate(
        self,
        on_ok: Callable[[Item, int | None], Sequence[ItemAlt]] = lambda x, y: [],
    ):
        self._on_ok = on_ok
        self.show()

    def _do_show(self):
        super()._do_show()
        self._search_entry.focus_set()
        self._on_query_entry_type()
        self._search_entry.select_range(0, "end")

    def _on_query_entry_type(self) -> None:
        self._item_list = self.item_source.search(self._search_query.get())
        self._result_list.set(cast(Any, [str(item) for item in self._item_list]))
        if self._item_list:
            self._result_listbox.select_clear(0, uni.END)
        # todo to clear or not to clear selection
        # if self._item_list:
        #     self._result_listbox.select_clear(0, uni.END)
        #     self._result_listbox.selection_set(0)
        self._result_listbox.see(0)


class Dialog(kvix.Dialog, ModalWindow):
    def __init__(self, parent: Ui, create_dialog: Callable[[kvix.DialogBuilder], None]):
        ModalWindow.__init__(self, parent)
        self.create_dialog = create_dialog
        self._init_window()

    def _init_window(self):
        self._window.bind("<Return>", cast(Any, self._ok_click))

        frame = uni.Frame(self._window, padding=8)
        frame.grid(column=0, row=0, sticky="nsew")
        frame.columnconfigure(0, weight=1)
        frame.rowconfigure(0, weight=1)

        data_frame = uni.Frame(frame)
        data_frame.grid(column=0, row=0, sticky="nsew")

        self.builder = DialogBuilder(data_frame)
        self.create_dialog(self.builder)
        children = data_frame.winfo_children()
        if len(children) >= 2:
            self._first_entry = children[1]

        control_frame = uni.Frame(frame)
        control_frame.rowconfigure(0, weight=1)
        control_frame.columnconfigure(0, weight=1)
        control_frame.grid(column=0, row=1, sticky="nsew")

        ok_button = uni.Button(control_frame, text=str(ok_text), command=self._ok_click)
        ok_button.grid(column=1, row=0, padx=4)

        cancel_button = uni.Button(control_frame, text=str(cancel_text), command=self.hide)
        cancel_button.grid(column=2, row=0, padx=4)

    def _ok_click(self, *args):
        self.hide()
        self.value = self.builder.save(self.value)
        self.on_ok()
        if self.auto_destroy:
            self.destroy()

    def _cancel_click(self, *args):
        self.hide()
        self.on_cancel()
        if self.auto_destroy:
            self.destroy()

    def activate(self) -> None:  # todo rename to "show"
        self.builder.load(self.value)
        self.show()

    def _do_show(self):
        super()._do_show()
        self._first_entry.select_range(0, "end")
        self._first_entry.focus_set()

    def destroy(self) -> None:
        return ModalWindow.destroy(self)


class DialogEntry(kvix.DialogEntry):
    def __init__(self, label: uni.Label, string_var: uni.StringVar):
        self._label = label
        self._string_var = string_var

    def set_title(self, text: str):
        self._label.config(text=text)

    def get_value(self):
        return self._string_var.get()

    def set_value(self, value):
        self._string_var.set(value)

    def on_change(self, func: callable):
        self._string_var.trace_add("write", lambda *args, **kwargs: func(self._string_var.get()))


class DialogBuilder(kvix.DialogBuilder):
    def __init__(self, root_frame: uni.Frame):
        super().__init__()
        self._root_frame = root_frame
        self._widget_count = 0

    def create_entry(self, id: str, title: str) -> DialogEntry:
        self._root_frame.columnconfigure(0, weight=1)

        label = uni.Label(self._root_frame, text=title)
        label.grid(column=0, row=self._widget_count, sticky="ew")
        self._widget_count += 1

        string_var = uni.StringVar(self._root_frame)
        entry = uni.Entry(self._root_frame, textvariable=string_var)
        entry.grid(column=0, row=self._widget_count, sticky="ew")
        self._widget_count += 1
        if self._widget_count == 2:
            entry.focus_set()

        separator = uni.Label(self._root_frame, text="")
        separator.grid(column=0, row=self._widget_count, sticky="ew")
        self._widget_count += 1

        return cast(DialogEntry, self._add_widget(id, DialogEntry(label, string_var)))



import kvix
from kvix import ActionType, Item
from kvix.impl import BaseActionType, BasePlugin, BaseItem
from kvix.plugin.builtin.machinist import BaseMachinist
from kvix.l10n import _
import re
import math
from kvix.util import apply_template
from typing import Sequence

action_type_title_text = _("Calculator").setup(ru_RU="Калькулятор", de_DE="Rechner")
action_description_text = _("Simple ariphmetic expression calculator").setup(
    ru_RU="Простой калькулятор", de_DE="Einfacher Rechner"
)
expression_error_text = _("Calculation error").setup(ru_RU="Ошибка вычисления", de_DE="Fehler")
item_title_text = _("= {{value}}")


class Action(BaseMachinist):
    def search(self, query: str) -> Sequence[Item]:
        if not query:
            return []
        if len(query) < 3:
            return []
        match = self._pattern and re.compile(self._pattern).match(query)
        result = self._evaluate(query)
        if not result:
            if match:
                result = str(expression_error_text)
            else:
                return []
        return [
            BaseItem(
                apply_template(str(item_title_text), {"value": result}),
                [
                    self._create_single_item_type_alt(lambda: result),
                    self._create_single_item_copy_alt(lambda: result),
                    self._create_single_item_paste_alt(lambda: result),
                ],
            )
        ]

    def _evaluate(self, query: str) -> str:
        math_functions = {key: getattr(math, key) for key in dir(math) if not key.startswith("_")}
        try:
            # todo work with commas instead of dots as fractional part separator
            # todo remove spaces between digits
            # todo remove commas as three digits separator
            return str(eval(query, math_functions, {}))
        except Exception:
            return ""


class Plugin(BasePlugin):
    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "kvix.plugin.builtin.calculator",
            str(action_type_title_text),  # " ".join(action_type_title.values()),
            action_factory=Action,
        )

    def get_actions(self) -> Sequence[kvix.Action]:
        return [
            Action(
                self._single_action_type,
                str(action_type_title_text),
                str(action_description_text),
                "(?i)^[0-9()\\+\\-\\*/ ]{3,}$",
            )
        ]



from kvix import ActionType, Context, DialogBuilder
from kvix.impl import BaseAction, BaseActionType, BasePlugin
from kvix.l10n import _
from kvix.conf import Conf, Item
import kvix
from typing import Any, Sequence

action_type_title_text = _("Settings").setup(ru_RU="Настройки", de_DE="Einstellungen")
default_action_title_text = action_type_title_text
default_action_description = " ".join(default_action_title_text.values())


class Action(BaseAction):
    def _run(self, **args: Any) -> None:
        def create_entry(builder: DialogBuilder, key: str, item: Item, title: str):
            entry = builder.create_entry(key, title)
            builder.on_load(lambda _: entry.set_value(str(item.read())))
            builder.on_save(lambda _: item.write(str(entry.get_value())))

        def create_dialog(
            builder: DialogBuilder,
            key_prefix: str = "",
            conf: Conf = self.action_type.context.conf,
            title_prefix: str = "",
        ):
            for key, scope in conf.scopes().items():
                create_dialog(
                    builder,
                    key_prefix + key + ".",
                    scope,
                    title_prefix + (scope.title + " | " if scope.title else ""),
                )
            for key, item in conf.items().items():
                create_entry(
                    builder,
                    key_prefix + key + ".",
                    item,
                    title_prefix + (item._title or item.key),
                )

        dialog = self.action_type.context.ui.dialog(create_dialog)
        dialog.on_ok = self.action_type.context.conf.save
        dialog.activate()


class Plugin(BasePlugin):
    def __init__(self, context: Context):
        super().__init__(context)

    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "kvix.plugin.builtin.settings",
            str(action_type_title_text),
            Action,
        )

    def get_actions(self) -> Sequence[kvix.Action]:
        return [
            Action(
                self._single_action_type,
                str(default_action_title_text),
                default_action_description,
            )
        ]



from __future__ import annotations
from typing import Sequence


import kvix
from kvix import ActionType
from kvix.impl import (
    BaseItem,
    BaseItemAlt,
    BaseAction,
    BaseActionType,
    BasePlugin,
    BaseItemSource,
)
from kvix.l10n import _
from kvix.util import query_match

add_action_text = _("Add Action").setup(ru_RU="Добавить действие", de_DE="Aktion hinzufuegen")
select_action_type_text = _("Select Action Type").setup(
    ru_RU="Выбор типа действия", de_DE="Aktionstyp auswählen"
)
select_text = _("Select").setup(ru_RU="Выбрать", de_DE="Auswählen")


class Action(BaseAction):
    def _run(self, query: str) -> None:
        action_type_selector = self.action_type.context.ui.selector()
        action_type_selector.title = str(select_action_type_text)

        def execute(action_type: ActionType) -> None:
            dialog = self.action_type.context.ui.dialog(action_type.create_editor)
            dialog.auto_destroy = True

            def on_ok() -> None:
                action_type_selector.destroy()
                if isinstance(dialog.value, kvix.Action):
                    self.action_type.context.action_registry.actions.append(dialog.value)
                    self.action_type.context.action_registry.save()
                # todo refresh main selector

            dialog.on_ok = on_ok
            dialog.activate()

        def search(query: str) -> Sequence[kvix.Item]:
            result: list[kvix.Item] = []
            for action_type in self.action_type.context.action_registry.action_types.values():

                def f(action_type: kvix.ActionType = action_type):
                    if not query or query_match(query, action_type.id, action_type.title):
                        result.append(
                            BaseItem(
                                action_type.title,
                                [
                                    BaseItemAlt(
                                        select_text,
                                        lambda: execute(action_type),
                                    )
                                ],
                            )
                        )

                f()
            return result

        action_type_selector.item_source = BaseItemSource(search)
        action_type_selector.activate()


class Plugin(BasePlugin):
    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "add-action",
            str(add_action_text),
            action_factory=Action,
        )

    def get_actions(self) -> Sequence[Action]:
        return [Action(self._single_action_type, str(add_action_text) + "...")]



from tkinter import Misc
from typing import Generator


def find_all_children(root: Misc) -> Generator[Misc, Misc, None]:
    yield root
    for item in root.winfo_children():
        yield from find_all_children(item)



from kvix import Context, ItemAlt
from kvix.impl import (
    BaseActionType,
    BasePlugin,
    BaseItemAlt
)
from kvix.l10n import _
import kvix
import webbrowser
from kvix.util import apply_template
from typing import Any, Sequence
from kvix.plugin.builtin.machinist import BaseMachinist

action_type_title_text = _("Open url").setup(ru_RU="Открыть ссылку")
item_title_text = _('Goto "{{url}}"').setup(ru_RU='Открыть "{{url}}"')
url_param_text = _("URL")

default_action_title_text = action_type_title_text
default_action_title_description = " ".join(default_action_title_text.values())


class Action(BaseMachinist):
    def _on_after_set_params(self, **params: Any) -> None:
        self._url = str(params["url"])

    def _create_single_item_alts(self, query: str) -> Sequence[ItemAlt]:
        return [
            BaseItemAlt(str(action_type_title_text), lambda: self._run(query)),
            self._create_single_item_type_alt(lambda: self._get_text(query)),
            self._create_single_item_copy_alt(lambda: self._get_text(query)),
            self._create_single_item_paste_alt(lambda: self._get_text(query)),
        ]

    def _get_text(self, query: str) -> str:
        return apply_template(self._url, {"query": query})

    def _run(self, query: str) -> None:
        self.action_type.context.ui.hide()
        webbrowser.open(self._get_text(query))


class Plugin(BasePlugin):
    def __init__(self, context: Context):
        super().__init__(context)

    def _create_single_action_type(self) -> kvix.ActionType:
        return BaseActionType(
            self.context,
            "kvix.plugin.builtin.openurl",
            str(action_type_title_text),
            Action,
            {"url": str(url_param_text)},
        )



import http.server
import http.client
import threading

import kvix.ui


PORT = 23844


def show_remote_widow(host, port):
    host = "127.0.0.1:" + PORT
    conn = http.client.HTTPSConnection(host)
    conn.request("GET", "/show-window")
    response = conn.getresponse()
    if response.status != 200:
        pass


def create_server(window: kvix.ui.Window):
    class Handler(http.server.BaseHTTPRequestHandler):
        def do_POST(self):
            if self.path == "/show-window":
                window.show()
                self.back(200, "ok")
            elif self.path == "/hide-window":
                window.hide()
                self.back(200, "ok")
            else:
                self.back(400, "wrong request")

        def back(self, status: int, msg: str):
            self.send_response(status)
            self.send_header("Content-Type", "application/json")
            self.end_headers()
            self.wfile.write(b'{"msg":"' + bytes(msg, "UTF-8") + b'"}')

    httpd = http.server.HTTPServer(("127.0.0.1", PORT), Handler)
    return threading.Thread(target=httpd.serve_forever)



from kvix import Action, ActionType
from kvix.impl import BaseAction, BaseActionType, BasePlugin
from kvix.l10n import _
from typing import Any, Sequence

quit_title_text = _("Quit kvix").setup(ru_RU="Выключить kvix", de_DE="kvix ausschalten")
quit_description = " | ".join([quit_title_text.default] + list(quit_title_text.l10ns.values()))


class QuitAction(BaseAction):
    def _run(self, **args: Any):
        self.action_type.context.quit()


class Plugin(BasePlugin):
    def _create_single_action_type(self) -> ActionType:
        return BaseActionType(
            self.context,
            "kvix.plugin.builtin.quit",
            str(quit_title_text),
            action_factory=QuitAction,
        )

    def get_actions(self) -> Sequence[Action]:
        return [
            QuitAction(
                self._single_action_type,
                str(quit_title_text),
                str(quit_description),
            )
        ]
